/* Copyright (C) 1989, 1992, 1993 Aladdin Enterprises.  All rights reserved.

This file is part of Ghostscript.

Ghostscript is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the Ghostscript General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
Ghostscript, but only under the conditions described in the Ghostscript
General Public License.  A copy of this license is supposed to have been
given to you along with Ghostscript so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.  */

/* gdevnpdl.c */
/*
 * NEC NPDL/NPDL(level 2) language printer driver for Ghostscript.
 *   Version 1.4
 *
 * This driver was designed to work with any NEC NPDL/NPDL(level 2)
 * language printers. The driver supports data compression with MH coding,
 * because computer-to-printer communication time is often a bottleneck.
 *
 * Any test reports, comments, suggestions, and bug reports are welcome.
 * 
 *  -- Osamu Watanabe <watanabe@bpe.es.osaka-u.ac.jp>
 *       Osaka University, Japan.
 */
#include "gdevprn.h"

private dev_proc_print_page(npdl_print_page);

/*
 * Valid values for IMAGE resolutions, that is, X_DPI and Y_DPI are
 * 160, 200, 240, and 400.
 * (Mr. Keita Kawabe <kawabe@granite.phys.s.u-tokyo.ac.jp> reported that
 *  the 600dpi printer, MultiWriter 2000X, accepts 600)
 * These two values must be the same.
 *
 * The resolution specified below is the default value.
 * You can also specify the resolution at runtime with -r option.
 */
#define X_DPI 400
#define Y_DPI 400

/*
 * Left, bottom, right, and top margins.
 *
 * The values defined below are more accurate for my printer,
 * and you had better to adjust these values to your printer.
 * The commented values are from the PC-PR1000E/4 User's Manual.
 */
#define L_MARGIN 0.35   /* 0.31 */
#define B_MARGIN 0.19   /* 0.20 */
#define R_MARGIN 0.13   /* 0.16 */
#define T_MARGIN 0.17   /* 0.20 */

/* The device descriptor */
gx_device_printer gs_npdl_device =
  prn_device(prn_std_procs, "npdl",
	DEFAULT_WIDTH_10THS_A4,                  /* default paper size: A4 */
	DEFAULT_HEIGHT_10THS_A4,
	X_DPI,                                   /* default resolution */
	Y_DPI,
	L_MARGIN, B_MARGIN, R_MARGIN, T_MARGIN,  /* margins */
	1, npdl_print_page);



/* ------ internal routines ------ */

/* modified from gdevpcl.h and gdevmjc.c */
#define PAPER_SIZE_LETTER 2
#define PAPER_SIZE_A5 25
#define PAPER_SIZE_A4 26
#define PAPER_SIZE_A3 27
#define PAPER_SIZE_B4 30
#define PAPER_SIZE_B5 31

/* terminating code */
private char
terminating[2][64][13]={{
/* white */
"00110101",     "000111",       "0111",         "1000",         /*  0 -  3 */
"1011",         "1100",         "1110",         "1111",         /*  4 -  7 */
"10011",        "10100",        "00111",        "01000",        /*  8 - 11 */
"001000",       "000011",       "110100",       "110101",       /* 12 - 15 */
"101010",       "101011",       "0100111",      "0001100",      /* 16 - 19 */
"0001000",      "0010111",      "0000011",      "0000100",      /* 20 - 23 */
"0101000",      "0101011",      "0010011",      "0100100",      /* 24 - 27 */
"0011000",      "00000010",     "00000011",     "00011010",     /* 28 - 31 */
"00011011",     "00010010",     "00010011",     "00010100",     /* 32 - 35 */
"00010101",     "00010110",     "00010111",     "00101000",     /* 36 - 39 */
"00101001",     "00101010",     "00101011",     "00101100",     /* 40 - 43 */
"00101101",     "00000100",     "00000101",     "00001010",     /* 44 - 47 */
"00001011",     "01010010",     "01010011",     "01010100",     /* 48 - 51 */
"01010101",     "00100100",     "00100101",     "01011000",     /* 52 - 55 */
"01011001",     "01011010",     "01011011",     "01001010",     /* 56 - 59 */
"01001011",     "00110010",     "00110011",     "00110100"      /* 60 - 63 */
},{
/* black */
"0000110111",   "010",          "11",           "10",           /*  0 -  3 */
"011",          "0011",         "0010",         "00011",        /*  4 -  7 */
"000101",       "000100",       "0000100",      "0000101",      /*  8 - 11 */
"0000111",      "00000100",     "00000111",     "000011000",    /* 12 - 15 */
"0000010111",   "0000011000",   "0000001000",   "00001100111",  /* 16 - 19 */
"00001101000",  "00001101100",  "00000110111",  "00000101000",  /* 20 - 23 */
"00000010111",  "00000011000",  "000011001010", "000011001011", /* 24 - 27 */
"000011001100", "000011001101", "000001101000", "000001101001", /* 28 - 31 */
"000001101010", "000001101011", "000011010010", "000011010011", /* 32 - 35 */
"000011010100", "000011010101", "000011010110", "000011010111", /* 36 - 39 */
"000001101100", "000001101101", "000011011010", "000011011011", /* 40 - 43 */
"000001010100", "000001010101", "000001010110", "000001010111", /* 44 - 47 */
"000001100100", "000001100101", "000001010010", "000001010011", /* 48 - 51 */
"000000100100", "000000110111", "000000111000", "000000100111", /* 52 - 55 */
"000000101000", "000001011000", "000001011001", "000000101011", /* 56 - 59 */
"000000101100", "000001011010", "000001100110", "000001100111"  /* 60 - 63 */
}};

/* make-up code */
private char
makeup[2][40][14]={{
/* white */
"11011",        "10010",        "010111",       "0110111",      /*   64- 256 */
"00110110",     "00110111",     "01100100",     "01100101",     /*  320- 512 */
"01101000",     "01100111",     "011001100",    "011001101",    /*  576- 768 */
"011010010",    "011010011",    "011010100",    "011010101",    /*  832-1024 */
"011010110",    "011010111",    "011011000",    "011011001",    /* 1088-1280 */
"011011010",    "011011011",    "010011000",    "010011001",    /* 1344-1536 */
"010011010",    "011000",       "010011011",    "00000001000",  /* 1600-1792 */
"00000001100",  "00000001101",  "000000010010", "000000010011", /* 1856-2048 */
"000000010100", "000000010101", "000000010110", "000000010111", /* 2112-2304 */
"000000011100", "000000011101", "000000011110", "000000011111"  /* 2368-2560 */
},{
/* black */
"0000001111",   "000011001000", "000011001001", "000001011011", /*   64- 256 */
"000000110011", "000000110100", "000000110101", "0000001101100",/*  320- 512 */
"0000001101101","0000001001010","0000001001011","0000001001100",/*  576- 768 */
"0000001001101","0000001110010","0000001110011","0000001110100",/*  832-1024 */
"0000001110101","0000001110110","0000001110111","0000001010010",/* 1088-1280 */
"0000001010011","0000001010100","0000001010101","0000001011010",/* 1344-1536 */
"0000001011011","0000001100100","0000001100101","00000001000",  /* 1600-1792 */
"00000001100",  "00000001101",  "000000010010", "000000010011", /* 1856-2048 */
"000000010100", "000000010101", "000000010110", "000000010111", /* 2112-2304 */
"000000011100", "000000011101", "000000011110", "000000011111"  /* 2368-2560 */
}};

/* EOL (end of line) code */
private char
eol[]="000000000001";

/* FILL bit */
private char
fill[]="0";

#define MAX_RUNLENGTH 2623  /* Max value of makeup(2560) + terminating(63) */
#define EOL_SIZE 12         /* Size of EOL (end of line) code */

private byte
mask[]={0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

/* Write MH code */
private int
write_mh_code(byte *out, int chunk_size, int num_bits, char *code)
{	int code_length, i, p, q;

	for(code_length = 0; code[code_length] != '\0'; code_length++);
	if(((num_bits + code_length) / 8) >= chunk_size) return 0;

	p = num_bits / 8;
	q = num_bits % 8;
	for(i = 0; i < code_length; i++){
		/*
		 * MH compressed data are stored from 2^0 to 2^7,
		 * whereas uncompressed data are stored from 2^7 to 2^0.
		 */
		if(code[i] == '0') out[p] &= ~(mask[7 - q]);
		else               out[p] |= mask[7 - q];

		if(q < 7) q++;
		else{ p++; q = 0; }
	}

	return code_length;
}

/* Set MH run-length code */
private int
set_mh_runlength(byte *out, int chunk_size, int num_bits, int phase, int count)
{	int n, code_length;

	code_length = 0;

	/* Set makeup code */
	if(count / 64 > 0){
		if((n = write_mh_code(out, chunk_size, num_bits,
			makeup[phase][(count / 64) - 1])) == 0) return 0;
		code_length += n;
	}

	/* Set terminating code */
	if((n = write_mh_code(out, chunk_size, num_bits + code_length,
		terminating[phase][count % 64])) == 0) return 0;
	code_length += n;

	return code_length;
}

/* Set MH EOL (end of line) code */
private int
set_mh_eol(byte *out, int chunk_size, int num_bits)
{
	return write_mh_code(out, chunk_size, num_bits, eol);
}

/* Set MH RTC (return to control) code and FILL bits */
private int
set_mh_rtc(byte *out, int chunk_size, int num_bits)
{	int i, n, code_length, num_fills;

	code_length = 0;

	/* Set FILL bits */
	num_fills = (EOL_SIZE * 6 + (num_bits % 8)) % 8;
	if(num_fills != 0){
		for(i = 0; i < 8 - num_fills; i++){
			if((n = write_mh_code(out, chunk_size,
				num_bits + code_length, fill)) == 0) return 0;
			code_length += n;
		}
	}

	/* Set RTC (that is, six EOL) code */
	for(i = 0; i < 6; i++){
		if((n = write_mh_code(out, chunk_size,
			num_bits + code_length, eol)) == 0) return 0;
		code_length += n;
	}

	return code_length;
}

/* Data compression by MH coding */
private int
mh_compression(byte *in, byte *out, int line_size, int column_size)
{	int i, p, q, r, n;
	int num_bits, phase, count;
	int chunk_size = line_size * column_size;
    	byte src;

	num_bits = 0;
	for(i = 0; i < column_size; i++){
		r = i * line_size;

		/* Synchronization of 1-dot line */
		if((n = set_mh_eol(out, chunk_size, num_bits)) == 0) return 0;
		num_bits += n;

		/* Data compression of 1-dot line */
		phase = count = 0;
		for(p = 0; p < line_size; p++){
			if(phase == 0) src = ~(in[r + p]);
			else           src = in[r + p];
			for(q = 0; q < 8; q++){

				/*
				 * If the dot has reversed, write the
				 * run-length of the continuous dots.
				 */
				if(!(src & mask[q])){
					if((n = set_mh_runlength(out,
						chunk_size, num_bits, phase,
						count)) == 0) return 0;
					num_bits += n;
					phase = (phase == 0) ? 1 : 0;
					count = 1;
					src = ~src;
				}
				/*
				 * If the dot is not reversed, check the
				 * length of this continuous dots,
				 */
				else{
					if(count < MAX_RUNLENGTH) count++;
					/*
					 * and if the length >= MAX_RUNLENGTH,
					 * stop and restart counting.
					 */
					else{
						if((n = set_mh_runlength(out,
							chunk_size, num_bits,
							phase, MAX_RUNLENGTH))
							== 0) return 0;
						num_bits += n;
		    				phase =	(phase == 0) ? 1 : 0;
						if((n = set_mh_runlength(out,
							chunk_size, num_bits,
							phase, 0))
							== 0) return 0;
						num_bits += n;
						phase = (phase == 0) ? 1 : 0;
						count = 1;
					}
				}
			}
		}

		/* Write the last run-length of 1-dot line */
		if((n = set_mh_runlength(out,chunk_size, num_bits, phase,
			count)) == 0) return 0;
		num_bits += n;
	}

	/* RTC (return to control) */
	if((n = set_mh_rtc(out, chunk_size, num_bits)) == 0) return 0;
	num_bits += n;

	return (num_bits / 8);
}

/* Get the paper size code, based on width and height. */
/* modified from gdevpcl.c and gdevmjc.c */
private int
gdev_npdl_paper_size(gx_device_printer *dev)
{	float height_inches = dev->height / dev->y_pixels_per_inch;
	return
	   (height_inches >= 15.9 ? PAPER_SIZE_A3 :
	    height_inches >= 13.6 ? PAPER_SIZE_B4 :
	    height_inches >= 11.1 ? PAPER_SIZE_A4 :
	    height_inches >= 10.4 ? PAPER_SIZE_LETTER :
	    height_inches >=  9.3 ? PAPER_SIZE_B5 :
	    PAPER_SIZE_A5);
}

/* Send the page to the printer.  For speed, compress each scan line, */
/* since computer-to-printer communication time is often a bottleneck. */
private int
npdl_print_page(gx_device_printer *pdev, FILE *prn_stream)
{	int line_size = gdev_prn_raster(pdev);
	int max_column_size = 16;
	int chunk_size = line_size * max_column_size;
	int bits_per_line = line_size * 8;
	int height = pdev->height;
	int x_dpi = (int)(pdev->x_pixels_per_inch);
	int paper_size;
	char paper_command[3];
	byte *in, *out;
	int lnum;

	/* Allocate memory */
	if((in = (byte *)gs_malloc(chunk_size, sizeof(byte),
		"npdl_print_page(in)")) == NULL)
		return_error(gs_error_VMerror);
	if((out = (byte *)gs_malloc(chunk_size, sizeof(byte),
		"npdl_print_page(out)")) == NULL){
		gs_free((char *)out, chunk_size, sizeof(byte),
			"npdl_print_page(out)");
		return_error(gs_error_VMerror);
	}

	/* Check paper size */
	paper_size = gdev_npdl_paper_size(pdev);
	switch(paper_size){
	  case PAPER_SIZE_A5:
		sprintf(paper_command, "A5"); break;
	  case PAPER_SIZE_A4:
		sprintf(paper_command, "A4"); break;
	  case PAPER_SIZE_A3:
		sprintf(paper_command, "A3"); break;
	  case PAPER_SIZE_B5:
		sprintf(paper_command, "B5"); break;
	  case PAPER_SIZE_B4:
		sprintf(paper_command, "B4"); break;
	  case PAPER_SIZE_LETTER:
		sprintf(paper_command, "LT"); break;
	}

	/* Initialize printer */
	fputs("\033c1", prn_stream);		  /* Software Reset */
	fputs("\034d240.", prn_stream);		  /* Page Printer Mode */
	sprintf(out, "\034fP%s.", paper_command); /* Paper Size */
	fputs(out, prn_stream);
	sprintf(out, "\034<1/%d,i.", x_dpi);	  /* Image Resolution */
	fputs(out, prn_stream);

	/* Send data to printer */
	for(lnum = 0; lnum < height;) {
		int column_size = 0;
		int i, code, num_bytes;

		/* Copy scan line */
		if((code=gdev_prn_copy_scan_lines(pdev, lnum, in, line_size))
			< 0) return_error(code);

		/* If the line is blank, continue */
		for(i = 0; i < line_size && *(in + i) == (byte)0; i++);
		if(i == line_size){
			lnum++;
			continue;
		}

		/* Send the upper-left coordinate of the chunk */
		sprintf(out, "\034e0,%d.", lnum);
		fputs(out, prn_stream);

		/* Copy chunk that have no blank line */
		for(lnum++, column_size++;
			lnum < height && column_size < max_column_size;
			lnum++, column_size++){
			byte *offseted = in + column_size * line_size;

			/* Copy next scan line */
			if((code=gdev_prn_copy_scan_lines(pdev, lnum, offseted,
				line_size)) < 0) return_error(code);

			/* If the line is blank, break */
			for(i = 0; i < line_size && *(offseted + i) == (byte)0;
				i++);
			if(i == line_size){
				lnum++;
				break;
			}
		}

		/* Data compression */
		num_bytes = mh_compression(in, out, line_size, column_size);

		/*
		 * If the compression ratio >= 100%, send uncompressed data
		 */
		if(num_bytes == 0){
			sprintf(out, "\034i%d,%d,0,1/1,1/1,%d,%d.",
				bits_per_line, column_size,
				line_size * column_size, x_dpi);
			fputs(out, prn_stream);
			fwrite(in, 1, line_size * column_size,
				prn_stream);
		}
		/*
		 * If the compression ratio < 100%, send compressed data
		 */
		else{
			sprintf(in, "\034i%d,%d,1,1/1,1/1,%d,%d.", 
				bits_per_line, column_size,
				num_bytes, x_dpi);
			fputs(in, prn_stream);
			fwrite(out, 1, num_bytes, prn_stream);
		}
	}

	/* Form Feed */
	fputs("\014", prn_stream);

	fflush(prn_stream);
	gs_free((char *)out, chunk_size, sizeof(byte), "npdl_print_page(out)");
	gs_free((char *)in, chunk_size, sizeof(byte), "npdl_print_page(in)");

	return 0;
}
